// Copyright 2018 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
package dev.facundo.fastqrreaderview.java;

import android.graphics.Bitmap;
import android.media.Image;
import android.util.Log;
import android.view.Surface;

import androidx.annotation.NonNull;

import com.google.android.gms.tasks.OnCompleteListener;
import com.google.android.gms.tasks.OnFailureListener;
import com.google.android.gms.tasks.OnSuccessListener;
import com.google.android.gms.tasks.Task;
import com.google.mlkit.vision.common.InputImage;

import java.nio.ByteBuffer;
import java.util.concurrent.atomic.AtomicBoolean;

import dev.facundo.fastqrreaderview.common.FrameMetadata;
import dev.facundo.fastqrreaderview.common.VisionImageProcessor;

/**
 * Abstract base class for ML Kit frame processors. Subclasses need to implement {@link
 * #onSuccess(T, FrameMetadata)} to define what they want to with the detection
 * results and {@link #detectInImage(InputImage)} to specify the detector object.
 *
 * @param <T> The type of the detected feature.
 */
public abstract class VisionProcessorBase<T> implements VisionImageProcessor {

    // Whether we should ignore process(). This is usually caused by feeding input data faster than
    // the model can handle.
    public final AtomicBoolean shouldThrottle = new AtomicBoolean(true);

    public VisionProcessorBase() {
    }

    @Override
    public void process(
            ByteBuffer data, final FrameMetadata frameMetadata) {
        if (shouldThrottle.get()) {
            return;
        }

        InputImage image = InputImage.fromByteBuffer(data.compact(), frameMetadata.getWidth(), frameMetadata.getHeight(), frameMetadata.getRotation(), InputImage.IMAGE_FORMAT_NV21);
        detectInVisionImage(image, frameMetadata);
//        Bitmap bitmap = Bitmap.createBitmap(frameMetadata.getWidth(), frameMetadata.getHeight(), Bitmap.Config.ARGB_8888);
//        data.rewind();
//        bitmap.copyPixelsFromBuffer(data);
//        byte[] arr = new byte[data.remaining()];
//        data.get(arr);
//        detectInVisionImage(InputImage.fromByteArray(arr, frameMetadata.getWidth(), frameMetadata.getHeight(), frameMetadata.getRotation(), InputImage.IMAGE_FORMAT_NV21), frameMetadata);
    }

    // Bitmap version
    @Override
    public void process(Bitmap bitmap) {//, final GraphicOverlay graphicOverlay) {
        if (shouldThrottle.get()) {
            return;
        }
        detectInVisionImage(InputImage.fromBitmap(bitmap, Surface.ROTATION_0), null);//, graphicOverlay);
    }

    /**
     * Detects feature from given media.Image
     *
     * @return created FirebaseVisionImage
     */
    @Override
    public void process(Image image, int rotation) {//, final GraphicOverlay graphicOverlay) {
        if (shouldThrottle.get()) {
            return;
        }

        // This is for overlay display's usage
        FrameMetadata frameMetadata =
                new FrameMetadata.Builder().setWidth(image.getWidth()).setHeight(image.getHeight
                        ()).build();
        InputImage fbVisionImage =
                InputImage.fromMediaImage(image, rotation);
        detectInVisionImage(fbVisionImage, frameMetadata);//, graphicOverlay);
    }

    private void detectInVisionImage(
            final InputImage image,
            final FrameMetadata metadata) {

        detectInImage(image)
                .addOnSuccessListener(
                        new OnSuccessListener<T>() {
                            @Override
                            public void onSuccess(T results) {
                                shouldThrottle.set(false);
                                VisionProcessorBase.this.onSuccess(results, metadata);//,
                                //graphicOverlay);
                            }
                        })
                .addOnFailureListener(
                        new OnFailureListener() {
                            @Override
                            public void onFailure(@NonNull Exception e) {
                                shouldThrottle.set(false);
                                VisionProcessorBase.this.onFailure(e);
                            }
                        });
        // Begin throttling until this frame of input has been processed, either in onSuccess or
        // onFailure.
        shouldThrottle.set(true);
    }

    @Override
    public void stop() {
    }

    protected abstract Task<T> detectInImage(InputImage image);

    protected abstract void onSuccess(
            @NonNull T results,
            @NonNull FrameMetadata frameMetadata);//,
//            @NonNull GraphicOverlay graphicOverlay);

    protected abstract void onFailure(@NonNull Exception e);
}
